@@ -34,6 +34,7 @@
   cd "$GOPATH"
   go get -u github.com/fatih/color
   go get -u github.com/fiam/gounidecode/unidecode
+  go get -u github.com/gedex/inflector
   go get -u github.com/klauspost/cpuid
   go get -u github.com/pbnjay/memory
   go get -u github.com/surgebase/porter2
@@ -80,6 +81,7 @@
 package main
 
 import (
+	"bytes"
 	"container/heap"
 	"fmt"
 	"golang.org/x/text/runes"
@@ -1255,13 +1257,16 @@ func RemoveEmbeddedMarkup(str string) string {
 	return buffer.String()
 }
 
-func HTMLAhead(text string, idx int) int {
+func HTMLAhead(text string, idx, txtlen int) int {
 
 	// record position of < character
 	start := idx
 
 	// at start of element
 	idx++
+	if idx >= txtlen {
+		return 0
+	}
 	ch := text[idx]
 
 	if ch == '/' {
@@ -1307,11 +1312,11 @@ func HTMLAhead(text string, idx int) int {
 	return idx + 1 - start
 }
 
-func HTMLBehind(bufr []byte, pos int) bool {
+func HTMLBehind(bufr []byte, pos, txtlen int) bool {
 
 	for pos >= 0 {
 		if bufr[pos] == '<' {
-			return HTMLAhead(string(bufr), pos) != 0
+			return HTMLAhead(string(bufr), pos, txtlen) != 0
 		}
 		pos--
 	}
@@ -1833,6 +1838,16 @@ func RepairScriptMarkup(str string, policy MarkupPolicy) string {
 	return buffer.String()
 }
 
+func RepairTableMarkup(str string, policy MarkupPolicy) string {
+
+	str = strings.Replace(str, "<tr>", " ", -1)
+	str = strings.Replace(str, "<td>", " ", -1)
+	str = strings.Replace(str, "</tr>", " ", -1)
+	str = strings.Replace(str, "</td>", " ", -1)
+
+	return str
+}
+
 func SplitInTwoAt(str, chr string, side SideType) (string, string) {
 
 	slash := strings.SplitN(str, chr, 2)
@@ -2220,6 +2235,7 @@ func CleanupContents(str string, ascii, amper, mixed bool) string {
 	if DoStrict {
 		if mixed || amper {
 			if HasAngleBracket(str) {
+				str = RepairTableMarkup(str, SPACE)
 				str = RemoveEmbeddedMarkup(str)
 			}
 		}
@@ -2282,6 +2298,11 @@ func DecodeFields(str string) string {
 			"[tiab]", " NORM ",
 			"[titl]", " NORM ",
 			"[year]", " YEAR ",
+			"[chem]", " CHEM ",
+			"[cebi]", " CEBI ",
+			"[disz]", " DISZ ",
+			"[gene]", " GENE ",
+			"[omim]", " OMIM ",
 		)
 	}
 
@@ -2336,10 +2357,10 @@ func CreateReader(in io.Reader) <-chan string {
 		Closed := false
 
 		// read one buffer, trim at last > and retain remainder for next call, signal if no > character
-		nextBuffer := func() (string, bool, bool) {
+		nextBuffer := func() ([]byte, bool, bool) {
 
 			if Closed {
-				return "", false, true
+				return nil, false, true
 			}
 
 			// prepend previous remainder to beginning of buffer
@@ -2347,7 +2368,7 @@ func CreateReader(in io.Reader) <-chan string {
 			Remainder = ""
 			if m > 16384 {
 				// previous remainder is larger than reserved section, write and signal need to continue reading
-				return string(Buffer[:m]), true, false
+				return Buffer[:m], true, false
 			}
 
 			// read next block, append behind copied remainder from previous read
@@ -2359,13 +2380,13 @@ func CreateReader(in io.Reader) <-chan string {
 					fmt.Fprintf(os.Stderr, "\nERROR: %s\n", err.Error())
 					// Ignore bytes - non-conforming implementations of io.Reader may returned mangled data on non-EOF errors
 					Closed = true
-					return "", false, true
+					return nil, false, true
 				}
 				// end of file
 				Closed = true
 				if n == 0 {
 					// if EOF and no more data, do not send final remainder (not terminated by right angle bracket that is used as a sentinel)
-					return "", false, true
+					return nil, false, true
 				}
 			}
 			if n < 0 {
@@ -2389,7 +2410,7 @@ func CreateReader(in io.Reader) <-chan string {
 				if bufr[pos] == '>' {
 					if DoStrict {
 						// optionally skip backwards past embedded i, b, u, sub, and sup HTML open, close, and empty tags, and MathML
-						if HTMLBehind(bufr, pos) {
+						if HTMLBehind(bufr, pos, len(bufr)) {
 							continue
 						}
 					}
@@ -2402,11 +2423,11 @@ func CreateReader(in io.Reader) <-chan string {
 			if pos > -1 {
 				pos++
 				Remainder = string(bufr[pos:])
-				return string(bufr[:pos]), false, false
+				return bufr[:pos], false, false
 			}
 
 			// no > found, signal need to continue reading long content
-			return string(bufr[:]), true, false
+			return bufr[:], true, false
 		}
 
 		// nextBlock reads buffer, concatenates if necessary to place long element content into a single string
@@ -2423,12 +2444,12 @@ func CreateReader(in io.Reader) <-chan string {
 
 			// if buffer does not end with > character
 			if cont {
-				var buff strings.Builder
+				var buff bytes.Buffer
 
 				// keep reading long content blocks
 				for {
-					if line != "" {
-						buff.WriteString(line)
+					if len(line) > 0 {
+						buff.Write(line)
 					}
 					if !cont {
 						// last buffer ended with sentinel
@@ -2442,28 +2463,28 @@ func CreateReader(in io.Reader) <-chan string {
 				}
 
 				// concatenate blocks
-				line = buff.String()
+				return buff.String()
 			}
 
-			// trimming spaces here would throw off line tracking
-
-			// optionally compress/cleanup tags/attributes and contents (undocumented)
-			if DoCleanup {
-				if HasBadSpace(line) {
-					line = CleanupBadSpaces(line)
-				}
-				if HasAdjacentSpaces(line) {
-					line = CompressRunsOfSpaces(line)
-				}
-			}
-
-			return line
+			return string(line)
 		}
 
 		// read XML and send blocks through channel
 		for {
 			str := nextBlock()
 
+			// trimming spaces here would throw off line tracking
+
+			// optionally compress/cleanup tags/attributes and contents (undocumented)
+			if DoCleanup {
+				if HasBadSpace(str) {
+					str = CleanupBadSpaces(str)
+				}
+				if HasAdjacentSpaces(str) {
+					str = CompressRunsOfSpaces(str)
+				}
+			}
+
 			out <- str
 
 			// bail after sending empty string sentinel
@@ -3109,7 +3130,7 @@ func ParseXML(Text, parent string, inp <-chan string, tokens func(Token), find *
 
 		if DoStrict && ch == '<' {
 			// check to see if an HTML or MathML element is at the beginning of a content string
-			if HTMLAhead(text, idx) != 0 {
+			if HTMLAhead(text, idx, txtlen) != 0 {
 				plainContent = false
 			}
 		}
@@ -3221,7 +3242,6 @@ func ParseXML(Text, parent string, inp <-chan string, tokens func(Token), find *
 							ch = text[idx]
 						}
 						str := text[start:idx]
-						ch = text[idx]
 						if ch != '>' {
 							// skip past unexpected blanks
 							for InBlank[ch] {
@@ -3373,10 +3393,12 @@ func ParseXML(Text, parent string, inp <-chan string, tokens func(Token), find *
 					}
 					if ch == '<' {
 						// optionally allow HTML text formatting elements and super/subscripts
-						advance := HTMLAhead(text, idx)
+						advance := HTMLAhead(text, idx, txtlen)
 						if advance > 0 {
 							idx += advance
-							ch = text[idx]
+							if idx < txtlen {
+								ch = text[idx]
+							}
 							plainContent = false
 							continue
 						}
@@ -4089,9 +4111,8 @@ func InitTables() {
 		InFirst[ch] = true
 	}
 	InFirst['_'] = true
-	// extend legal initial letter with at sign to handle biological data converted from JSON
+	// extend legal initial letter with at sign and digits to handle biological data converted from JSON
 	InFirst['@'] = true
-	// extend legal initial letter with digits to handle biological data converted from JSON
 	for ch := '0'; ch <= '9'; ch++ {
 		InFirst[ch] = true
 	}
@@ -4113,8 +4134,6 @@ func InitTables() {
 	InElement['-'] = true
 	InElement['.'] = true
 	InElement[':'] = true
-	// extend legal element letters with plus sign to handle biological data converted from JSON
-	InElement['+'] = true
 
 	// embedded markup and math tags are lower case
 	for i := range InLower {
